selfDir_uZPPqC := $(dir $(lastword $(MAKEFILE_LIST)))

# Make
SHELL := /bin/bash
subMake = $(MAKE) $(@F) --directory=$(@D)
+_ = $(+:_%=)
addBeforeExt = $(basename $(2))$(1)$(suffix $(2))

# Terminal
esc := '['
reset := $(esc)'0m'
emph := $(esc)'7m '
endEmph := ' '$(reset)

# Commands
CP = cp -p
DIFF = diff --unified=2

# Config
test_n ?= 2
tablesSort ?= plots organisms stems

+maps = $(filter maps/%,$(+_))
+in = $(filter-out maps/%,$(+_))

root := $(selfDir_uZPPqC)..
psqlAsBien := $(root)/bin/psql_vegbien

#####

all: _always maps ;

.SUFFIXES:

_always:
.PHONY: _always

clean: _always
	$(RM) $(all)

remake: _always clean all ;

%.out: %.make _always
	./$* >$@
.PRECIOUS: %.out

$(root)/%: _always
	+$(subMake)

#####

reinstall: _always uninstall install ;

#####

allViaMaps := $(filter-out maps/VegBIEN.%.csv,$(wildcard maps/*.csv))
via := $(firstword $(sort $(basename $(basename $(notdir $(allViaMaps))))))
viaMapsSort := $(wildcard $(tablesSort:%=maps/$(via).%.csv))
viaMaps := $(viaMapsSort) $(filter-out $(viaMapsSort),$(viaMaps))
tables := $(viaMaps:maps/$(via).%.csv=%)
autogenMaps := $(subst $(via).,VegBIEN.,$(viaMaps))
directMaps := $(autogenMaps) $(filter-out $(autogenMaps),\
$(wildcard maps/VegBIEN.*.csv))

maps: $(directMaps) _always ;

.PRECIOUS: $(allViaMaps) $(directMaps)

all += $(autogenMaps)

maps/VegBIEN.%.csv: maps/$(via).%.csv $(root)/mappings/$(via)-VegBIEN.%.csv
	$(root)/bin/join_sort <$+ >$@
.PRECIOUS: maps/VegBIEN.%.csv $(root)/mappings/$(via)-VegBIEN.%.csv

#####

dbFile := $(firstword $(wildcard src/db.*.sql))

ifneq ($(wildcard $(dbFile)),)

import: _always import-all verify ;

log = $(@:-all=)$(if $(n),.n=$(n),).log
import = -(set -x; "time" env commit=1 $(map2db) $(+_)) \
$(if $(n),,>>$(log)) 2>&1$(if $(n),|tee -a $(log))
# ignore import errors, which are often benign (e.g. invalid date format)

import-all: $(directMaps) _always
	$(import)

import-%: maps/VegBIEN.%.csv _always
	$(import)

logs := $(wildcard *.log)

rm_logs: _always
	$(RM) $(logs)

else
import: _always ;
import-%: _always ;
rm_logs: _always ;
endif

#####

ifneq ($(wildcard verify.ref*),)

verify: verify.ref verify.out _always
	$(DIFF) $(+_)

all += verify.out

%.out: %.sql _always
	$(out_cmd)
.PRECIOUS: %.out
out_cmd = $(psqlAsBien) --no-align --field-separator='	' --pset=footer=off \
--pset=null=NULL <$< >$@

else
verify: _always ;
endif

#####

ifneq ($(wildcard test/),)

tests :=

hasOwnRef = $(filter-out %.2-step.xml,$@)
testRef = $(1:.2-step.xml=.xml).ref

define test
env test=1 verbose=1 n=$(test_n) $(1) $(if $(+in),<$(+in)) $(+maps) >$@
@(set -x; $(DIFF) $(call testRef,$@) $@) 2>&1 || { e=$$?;\
$(if $(hasOwnRef),echo $(emph)"To accept new test output:"$(endEmph);\
echo "$(MAKE) $@-ok --directory=$(realpath .) --makefile=../input.Makefile";)\
exit $$e;}
endef

test/$(via).xml: $(viaMaps) _always
	$(call test,$(map))
tests += test/$(via).xml

test/VegBIEN.xml: $(directMaps) _always
	$(call test,$(map))
tests += test/VegBIEN.xml

test/VegBIEN.2-step.xml: test/$(via).xml\
$(root)/mappings/$(via)-VegBIEN.organisms.csv _always
	-$(call test,$(root)/map)
# Don't abort tester if only 2-step test fails, as it's often finicky
tests += test/VegBIEN.2-step.xml

test/import.out: $(directMaps) _always
	$(call test,$(map2db))
tests += test/import.out

test: _always $(tests) ;

.PRECIOUS: $(tests)

all += $(tests)

# Accepts a test output: make <test_output_path>-ok
%-ok: _always
	$(CP) $* $(call testRef,$*)

else
test: _always ;
endif

#####

# Each input type needs var $(mapEnv) and targets install, uninstall

ifneq ($(dbFile),)

dbEngineExt := $(subst .,,$(suffix $(basename $(notdir $(dbFile)))))
db := $(notdir $(realpath .))

%.ref: %.ref.sql
	$(inputDbAsBien) $(db) <$< >$@
.PRECIOUS: %.ref

####

install: _always db ;

uninstall: _always rm_db ;

####

# Each DB engine needs vars $(dbEngine), $(dbAsBien) and targets db, rm_db

ifeq ($(dbEngineExt),my)

dbEngine := MySQL

bienPassword := $(shell cat $(root)/config/bien_password)
mysqlAs = mysql --user=$(1) --password='$(bienPassword)'
mysqlAsRoot := $(call mysqlAs,root)
dbAsBien := $(call mysqlAs,bien)

db: $(dbFile) _always
	echo "CREATE DATABASE $(db) DEFAULT CHARACTER SET latin1;"|$(mysqlAsRoot)
	-$(mysqlAsRoot) --database=$(db) <$<
	echo "GRANT SELECT ON $(db).* TO 'bien'@'localhost';"|$(mysqlAsRoot)
# ignore errors in db import so that GRANT will still be run

rm_db: _always
	-echo "REVOKE ALL ON $(db).* FROM 'bien'@'localhost';"|$(mysqlAsRoot)
	echo "DROP DATABASE IF EXISTS $(db);"|$(mysqlAsRoot)
# ignore errors if grant not defined

###

else
$(error The DB filename $(dbFile) must have the form db.{my|pg}.sql)
endif

# Must come after dbEngine is set
mapEnv := in_engine=$(dbEngine) in_database=$(db)

####

else
mapEnv :=
install: _always ;
uninstall: _always ;
endif

#####

# Must come after mapEnv is set
map := $(if $(mapEnv),env $(mapEnv) ,)$(root)/map
map2db := env out_database=vegbien $(map)
